#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>

#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <wiringPi.h>

#include "max7219.h"
#include "SMS_UDP_Definitions.hpp"

int sockfd_udp_cli;

void intHandler(int dummy)
{
	close(sockfd_udp_cli);
	send_MAX7219(SHUTDOWN, 0, 0);
	exit(0);
}

int main(int argc, char** argv)
{
	signal(SIGINT, intHandler);
	
	if( wiringPiSetup() < 0 )
	{
		return 1;
	}
	
	puts("\nLocoField SLI\n");

	char udp_recvbuff[SMS_UDP_MAX_PACKETSIZE];
	PacketBase packetHeader;
	memset(&packetHeader, 0, sizeof(packetHeader));

	sockfd_udp_cli = socket(AF_INET, SOCK_DGRAM, 0);
	if( sockfd_udp_cli < 0 )
	{
		puts("ERROR: create socket.\n");
		return 1;
	}
	
	//int flags = fcntl(sockfd_udp_cli, F_GETFL, 0);
	//fcntl(sockfd_udp_cli, F_SETFL, flags | O_NONBLOCK);
	
	struct timeval timeout;      
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;
	setsockopt(sockfd_udp_cli, SOL_SOCKET, SO_RCVTIMEO, (char*)&timeout, sizeof(timeout));

	
	sockaddr_in receiver;
	receiver.sin_family = AF_INET;
	receiver.sin_addr.s_addr = INADDR_ANY;
	receiver.sin_port = htons(SMS_UDP_PORT);

	int retval = bind(sockfd_udp_cli, (sockaddr*) &receiver, sizeof(receiver));
	if( retval < 0 )
	{
		puts("ERROR: bind socket.\n");
		return 1;
	}
	
	
	pinMode(DATA, OUTPUT);
	pinMode(CLOCK, OUTPUT);
	pinMode(LOAD, OUTPUT);

	send_MAX7219(SCAN_LIMIT, 7, 7);
	send_MAX7219(DECODE_MODE, 0, 0);
	send_MAX7219(INTENSITY, 8, 8);
	send_MAX7219(SHUTDOWN, 1, 1);
	
	int digit_table[11] = { 126, 48, 109, 121, 51, 91, 95, 114, 127, 123, 0 };
	int dot_table[9][9] =
	{ // 128 64 32 16 8 4 2 1
	{ 0, 0, 0, 0, 15, 0, 0, 0, 0 }, // N
	{ 0, 1, 1, 1, 1, 1, 1, 1, 0 }, // 1
	{ 0, 15, 1, 1, 15, 8, 8, 15, 0 }, // 2
	{ 0, 15, 1, 1, 15, 1, 1, 15, 0 }, // 3
	{ 0, 9, 9, 9, 15, 1, 1, 1, 0 }, // 4
	{ 0, 15, 8, 8, 15, 1, 1, 15, 0 }, // 5
	{ 0, 15, 8, 8, 15, 9, 9, 15, 0 }, // 6
	{ 0, 15, 9, 9, 9, 1, 1, 1, 0 }, // 7
	{ 0, 0, 0, 0, 15, 8, 8, 8, 0 }, // R
	};
	
	send_MAX7219(1, digit_table[0], dot_table[0][1]);
	send_MAX7219(2, digit_table[0], dot_table[0][2]);
	send_MAX7219(3, digit_table[0], dot_table[0][3]);
	send_MAX7219(4, digit_table[0], dot_table[0][4]);
	send_MAX7219(5, digit_table[0], dot_table[0][5]);
	send_MAX7219(6, digit_table[0], dot_table[0][6]);
	send_MAX7219(7, digit_table[0], dot_table[0][7]);
	send_MAX7219(8, digit_table[0], dot_table[0][8]);
	
	
	sTelemetryData data1;
	memset(&data1, 0, sizeof(data1));
	
	while( 1 )
	{
		retval = recvfrom(sockfd_udp_cli, (char*)&udp_recvbuff, sizeof(udp_recvbuff), 0, NULL, 0);	
		if( retval > 0 )
		{
			memcpy(&packetHeader, udp_recvbuff, sizeof(PacketBase));
			switch( packetHeader.mPacketType )
			{
			case eCarPhysics:
			{
				memcpy(&data1, udp_recvbuff, sizeof(data1));
				
				// speed
				int speed = data1.sSpeed / 1000 * 3600;
				int s3 = speed / 100; speed -= s3 * 100;
				int s2 = speed / 10; speed -= s2 * 10;
				int s1 = speed;
				
				// rpm
				int rpm = data1.sRpm;
				int r5 = 10;
				int r4 = rpm / 1000; rpm -= r4 * 1000;
				int r3 = rpm / 100; rpm -= r3 * 100;
				int r2 = rpm / 10; rpm -= r2 * 10;
				int r1 = rpm;
				
				if( r4 >= 10 )
				{
					r4 -= 10;
					r5 = 1;
				}
				
				int gear = data1.sGearNumGears;
				int maxGear = gear >> 4;
				int curGear = gear & 15;
				
				if( curGear == 15 ) // reverse gear;
					curGear = 8;
				
				float rpm_per = (float) data1.sRpm / data1.sMaxRpm - 0.8;
				int rpmL = 0;
				
				if( rpm_per > 0 )
					rpmL = rpm_per /= 0.025;
				
				int led[9] =
				{
					0,
					digit_table[r1],
					digit_table[r2],
					digit_table[r3],
					digit_table[r4],
					digit_table[r5],
					digit_table[s1] | 128,
					digit_table[s2],
					digit_table[s3],
				};
				
				for( int i = 8; i > 0; i-- )
				{
					int dot = dot_table[curGear][i];
					
					if( rpmL != 0 )
					{
						dot |= 128;
						rpmL--;
					}
					
					send_MAX7219(i, led[i], dot);
				}

				usleep(1);
			}
			default:
				continue;
			}
		}
	}

	intHandler(0);
	return 0;
}
